@@ -1,9 +1,23 @@
 from __future__ import annotations
 
-from pathlib import Path
-from typing import Any, Dict, Generic, Iterator, Literal, Optional, Type, TypeVar, Union, get_args
+from pathlib import Path, PurePath
+from tempfile import TemporaryDirectory
+from typing import (
+    Any,
+    Dict,
+    Generic,
+    Iterator,
+    Literal,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    get_args,
+)
 from urllib.parse import unquote
 
+from openpyxl import load_workbook
 from pydantic import (
     AnyHttpUrl,
     BaseModel,
@@ -150,16 +164,19 @@ class ApiModel(BaseModel):
     @classmethod
     def from_id(cls: Type[ModelType], id: int, **kwargs) -> ModelType:
         session = current_session()
+        cache = session.cache[cls]
 
-        r: requests.Response = session.get(f'{cls._endpoint}/{id}/')
-        raise_for_status(r)
-        json = r.json()
-        for key, value in cls.__fields__.items():
-            if key in kwargs:
-                json[key] = kwargs[key]
-            elif value.type_ is not Any and issubclass(value.type_, ApiModel):
-                json[key] = value.type_.from_id(json[key])
-        return cls(**json)
+        if id not in cache:
+            r: requests.Response = session.get(f'{cls._endpoint}/{id}/')
+            raise_for_status(r)
+            json = r.json()
+            for key, value in cls.__fields__.items():
+                if key in kwargs:
+                    json[key] = kwargs[key]
+                elif value.type_ is not Any and issubclass(value.type_, ApiModel):
+                    json[key] = value.type_.from_id(json[key])
+            cache[id] = cls(**json)
+        return cache[id]
 
     @classmethod
     def list(cls: Type[ModelType], **kwargs) -> Iterator[ModelType]:
@@ -246,6 +263,26 @@ class Dataset(ApiModel):
     def add_subject(self, name: str) -> Subject:
         return Subject(name=name, dataset=self).create()
 
+    @property
+    def projects(self) -> Iterator[Project]:
+        self.assert_remote()
+        return Project.list(dataset=self)
+
+    @property
+    def segmentations(self) -> Iterator[Segmentation]:
+        for subject in self.subjects:
+            for segmentation in subject.segmentations:
+                yield segmentation
+
+    def add_project(self, file: Path, keywords: str = '', description: str = '') -> Project:
+        project = Project(
+            file=file,
+            keywords=keywords,
+            description=description,
+            dataset=self,
+        ).create()
+        return project.load_project_spreadsheet()
+
     @classmethod
     def from_name(cls, name: str) -> Optional[Dataset]:
         results = cls.list(name=name)
@@ -254,6 +291,46 @@ class Dataset(ApiModel):
         except StopIteration:
             return None
 
+    def load_data_spreadsheet(self, file: Union[Path, str]) -> Dataset:
+        file = Path(file)
+        xls = load_workbook(str(file), read_only=True)
+        if 'data' not in xls:
+            raise Exception('`data` sheet not found')
+
+        data = xls['data'].values
+
+        if next(data)[0] != 'shape_file':
+            raise Exception('Unknown spreadsheet format')
+
+        root = file.parent
+        subjects: Dict[str, Subject] = {}
+
+        for row in data:
+            shape_file = root / row[0]
+            if not shape_file.exists():
+                raise Exception(f'Could not find shape file at "{shape_file}"')
+
+            subject_name = shape_file.stem
+            if subject_name not in subjects:
+                subjects[subject_name] = self.add_subject(subject_name)
+
+            subject = subjects[subject_name]
+
+            # TODO: where to find the anatomy type?
+            subject.add_segmentation(file=shape_file, anatomy_type='unknown')
+
+        return self
+
+    def download(self, path: Union[Path, str]):
+        self.assert_remote()
+        path = Path(path)
+        for segmentation in self.segmentations:
+            # TODO: add a dataset spreadsheet to the data model and get the path from it
+            segmentation.file.download(path / 'segmentations')
+
+        for project in self.projects:
+            project.download(path)
+
 
 class Subject(ApiModel):
     _endpoint = 'subjects'
@@ -284,6 +361,7 @@ class Project(ApiModel):
     file: FileType[Literal['core.Project.file']]
     keywords: str = ''
     description: str = ''
+    dataset: Dataset
 
     @property
     def groomed_segmentations(self) -> Iterator[GroomedSegmentation]:
@@ -316,6 +394,116 @@ class Project(ApiModel):
             parameters=parameters,
         ).create()
 
+    def load_project_spreadsheet(self) -> Project:
+        file = self.file.path
+        assert file  # should be guaranteed by assert_local
+
+        root = file.parent
+        xls = load_workbook(str(file), read_only=True)
+
+        segmentations = {
+            PurePath(segmentation.file.name).stem: segmentation
+            for segmentation in self.dataset.segmentations
+        }
+
+        if 'optimize' not in xls or 'data' not in xls:
+            raise Exception('`data` sheet not found')
+
+        shape_model = self._parse_optimize_sheet(xls['optimize'].values)
+        self._parse_data_sheet(segmentations, shape_model, xls['data'].values, root)
+
+        return self
+
+    def _iter_data_sheet(
+        self, sheet: Any, root: Path
+    ) -> Iterator[Tuple[Path, Path, str, Path, Path]]:
+        headers = next(sheet)
+        if headers != (
+            'shape_file',
+            'groomed_file',
+            'alignment_file',
+            'local_particles_file',
+            'world_particles_file',
+        ):
+            raise Exception('Unknown spreadsheet format')
+
+        for row in sheet:
+            shape_file, groomed_file, alignment_file, local, world = row
+
+            shape_file = root / shape_file
+            groomed_file = root / groomed_file
+            local = root / local
+            world = root / world
+
+            yield shape_file, groomed_file, alignment_file, local, world
+
+    def _parse_optimize_sheet(self, sheet: Any) -> OptimizedShapeModel:
+        headers = next(sheet)
+        if headers != ('key', 'value'):
+            raise Exception('Unknown spreadsheet format')
+
+        params: Dict[str, Union[str, float]] = {}
+        for row in sheet:
+            key, value = row
+            try:
+                value = float(value)
+            except ValueError:
+                pass
+            params[key] = value
+
+        return self.add_shape_model(params)
+
+    def _parse_data_sheet(
+        self,
+        segmentations: Dict[str, Segmentation],
+        shape_model: OptimizedShapeModel,
+        sheet: Any,
+        root: Path,
+    ):
+        for shape_file, groomed_file, alignment_file, local, world in self._iter_data_sheet(
+            sheet, root
+        ):
+            segmentation = segmentations.get(shape_file.stem)
+            if not segmentation:
+                raise Exception(f'Could not find segmentation for "{shape_file}"')
+
+            groomed_segmentation = self.add_groomed_segmentation(
+                file=groomed_file,
+                segmentation=segmentation,
+            )
+            with TemporaryDirectory() as dir:
+                transform = Path(dir) / f'{shape_file.stem}.transform'
+                with transform.open('w') as f:
+                    f.write(alignment_file)
+
+                shape_model.add_particles(
+                    world=world,
+                    local=local,
+                    transform=transform,
+                    groomed_segmentation=groomed_segmentation,
+                )
+
+    def download(self, path: Union[Path, str]):
+        path = Path(path)
+        project_file = self.file.download(path)
+        xls = load_workbook(str(project_file), read_only=True)
+        sheet = xls['data'].values
+
+        shape_model = next(self.shape_models)
+        groomed_segmentations = {
+            PurePath(gs.file.name).stem: gs for gs in self.groomed_segmentations
+        }
+        local_files = {PurePath(p.local.name).stem: p for p in shape_model.particles}
+
+        # TODO: Do we detect if alignment_file (transform) is a path?
+        for _, groomed_file, _, local, world in self._iter_data_sheet(sheet, path):
+            gs = groomed_segmentations[groomed_file.stem]
+            gs.file.download(groomed_file.parent)
+
+            particles = local_files[local.stem]
+            particles.local.download(local.parent)
+            particles.world.download(world.parent)
+
 
 class GroomedSegmentation(ApiModel):
     _endpoint = 'groomed-segmentations'
